//--------------------------------------------------------------------------------------
// Calculate cubic splines from a set of control points/knots
// and performs cubic interpolation
//
// DirectX/HLSL implementation by
// Christian B. Mendl, 2012
//
// Reference:
// http://graphicsrunner.blogspot.de/2009/01/volume-rendering-102-transfer-functions.html
//--------------------------------------------------------------------------------------

#ifndef CUBIC_SPLINE_H
#define CUBIC_SPLINE_H

#include <d3d11.h>
#include <xnamath.h>


/// Rows hold the a, b, c and d coefficients (4 floats each)
typedef XMMATRIX SplineCoeffs;


struct ControlPoint
{
	XMVECTOR coord;
	UINT isoval;
};


// Evaluate the point using a cubic equation
inline XMVECTOR GetPointOnSpline(const SplineCoeffs *c, float s)
{
	return (((c->r[3] * s) + c->r[2]) * s + c->r[1]) * s + c->r[0];
}


//--------------------------------------------------------------------------------------
// Calculate spline coefficients by solving symmetric tridiagonal system
//
// Input:	cpoints		control points: there are (n+1) number of
//						'cpoints' i = 0, 1, ..., n
//
// Output:	C			memory of result vector C must have been allocated before
//						function call (n * sizeof(SplineCoeffs))
//--------------------------------------------------------------------------------------
void CalculateSplineCoeffs(int n, const ControlPoint *cpoints, SplineCoeffs *C);


//--------------------------------------------------------------------------------------
// Fit a cubic spline to the control points and evaluate at iso-values
//
// Input:	cpoints		control points: there are (n+1) number of
//						'cpoints' i = 0, 1, ..., n
//
// Output:	coords		result vector of length cpoints[n].isoval - cpoints[0].isoval
//--------------------------------------------------------------------------------------
void SplineInterpolate(int n, const ControlPoint *cpoints, XMVECTOR *coords);



#endif
