//--------------------------------------------------------------------------------------
// Volume ray-casting main algorithm
//
// DirectX/HLSL implementation by
// Christian B. Mendl, 2012
//
// Reference:
// http://graphicsrunner.blogspot.de/2009/01/volume-rendering-101.html
//--------------------------------------------------------------------------------------


#pragma pack_matrix(row_major)


//--------------------------------------------------------------------------------------
// Textures and samplers
//--------------------------------------------------------------------------------------

Texture3D<float4> txGradientVoxel : register(t0);

Texture2D<float4> txPositionFront : register(t1);
Texture2D<float4> txPositionBack  : register(t2);

Texture1D<float4> txTransfer      : register(t3);

SamplerState samplerLinear : register(s0);


//--------------------------------------------------------------------------------------
// Constants and constant buffer variables
//--------------------------------------------------------------------------------------

static const uint g_iMaxIterations = 512;

// Diagonal of a unit cube has length sqrt(3)
static const float g_fStepSize = sqrt(3.f)/g_iMaxIterations;

// for vertex shader
cbuffer cbEveryFrameVS : register(b0)
{
	matrix mWVP;
}

// for pixel shader

cbuffer cbImmutablePS : register(b0)
{
	float2 g_fInvWindowSize;
}

cbuffer cbEveryFramePS : register(b1)
{
	float4 vPosLight;
}


//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------

struct VSInput
{
	float4 pos : POSITION;
};

struct PSInput
{
	float4 pos : SV_POSITION;
};


//--------------------------------------------------------------------------------------
// Vertex shaders
//--------------------------------------------------------------------------------------

PSInput RayCastVS(VSInput input)
{
	PSInput output;
	output.pos = mul(mWVP, input.pos);

	return output;
}


//--------------------------------------------------------------------------------------
// Pixel shaders
//--------------------------------------------------------------------------------------

float4 RayCastPS(PSInput input) : SV_TARGET
{
	// Current pixel location on screen, used to sample position texture
	float2 tex = input.pos.xy * g_fInvWindowSize;
 
	// Read cube front and back face positions (in model coordinates) from texture
	float3 pos_front = (float3)txPositionFront.Sample(samplerLinear, tex);
	float3 pos_back  = (float3) txPositionBack.Sample(samplerLinear, tex);
 
	// Direction vector
	float3 dir = normalize(pos_back - pos_front);

	// Number of iterations
	//uint numIter = min(ceil(ray_length/g_fStepSize), g_iMaxIterations);

	// Single step: direction times delta step
	float3 step = g_fStepSize * dir;

	// Current position
	float3 v = pos_front;

	// Accumulate result: value and transparency (alpha)
	float4 result = float4(0, 0, 0, 0);
 
	for (uint i = 0; i < /*numIter*/g_iMaxIterations; i++)
	{
		// sample gradient and iso-value
		float4 gradIso = txGradientVoxel.Sample(samplerLinear, v);
		
		// Index the transfer function with the iso-value and get the rgba value for the voxel
		float4 src = txTransfer.Sample(samplerLinear, gradIso.a);

		// Diffuse lighting
		float s = clamp(dot(gradIso.xyz, vPosLight.xyz), 0, 1);
				
		// Diffuse shading and fake ambient lighting
		src.rgb = s * src.rgb + 0.1f * src.rgb;

		// Front to back blending
		src.rgb *= src.a;
		result += (1 - result.a) * src;

		// Advance the current position
		v += step;
	}
 
	//return float4(result.r, result.r, result.r, result.y);
	return result;
}
