//--------------------------------------------------------------------------------------
// DirectX/HLSL implementation by
// Christian B. Mendl, 2012
//
// Reference:
// http://graphicsrunner.blogspot.de/2009/01/volume-rendering-102-transfer-functions.html
//--------------------------------------------------------------------------------------

#include "cubic_spline.h"
#include <assert.h>


//--------------------------------------------------------------------------------------
// Calculate spline coefficients by solving symmetric tridiagonal system
//--------------------------------------------------------------------------------------
void CalculateSplineCoeffs(int n, const ControlPoint *v, SplineCoeffs *C)
{
	int i;

	XMVECTOR *gamma = (XMVECTOR *)malloc((n + 1)*sizeof(XMVECTOR));
	XMVECTOR *delta = (XMVECTOR *)malloc((n + 1)*sizeof(XMVECTOR));
	XMVECTOR *D     = (XMVECTOR *)malloc((n + 1)*sizeof(XMVECTOR));

	//	We need to solve the equation
	//	taken from: http://mathworld.wolfram.com/CubicSpline.html
	//	|2 1       | |D[0]|   |3(v[1] - v[0])  |
	//	|1 4 1     | |D[1]|   |3(v[2] - v[0])  |
	//	|  1 4 1   | | .  | = |      .         |
	//	|    ..... | | .  |   |      .         |
	//	|     1 4 1| | .  |   |3(v[n] - v[n-2])|
	//	|       1 2| |D[n]|   |3(v[n] - v[n-1])|
	//
	//	by converting the matrix to upper triangular.
	//	The D[i] are the derivatives at the control points.

	XMVECTOR one = XMVectorSet(1, 1, 1, 1);

	// This builds the coefficients of the left matrix
	gamma[0] = XMVectorSet(0.5f, 0.5f, 0.5f, 0.5f);
	for (i = 1; i < n; i++)
	{
		gamma[i] = one / ((4 * one) - gamma[i - 1]);
	}
	gamma[n] = one / ((2 * one) - gamma[n - 1]);

	delta[0] = 3 * (v[1].coord - v[0].coord) * gamma[0];
	for (i = 1; i < n; i++)
	{
		delta[i] = (3 * (v[i + 1].coord - v[i - 1].coord) - delta[i - 1]) * gamma[i];
	}
	delta[n] = (3 * (v[n].coord - v[n - 1].coord) - delta[n - 1]) * gamma[n];

	D[n] = delta[n];
	for (i = n - 1; i >= 0; i--)
	{
		D[i] = delta[i] - gamma[i] * D[i + 1];
	}

	// Compute the coefficients of the cubics
	for (i = 0; i < n; i++)
	{
		C[i] = SplineCoeffs(v[i].coord, D[i], 3 * (v[i + 1].coord - v[i].coord) - 2 * D[i] - D[i + 1], 2 * (v[i].coord - v[i + 1].coord) + D[i] + D[i + 1]);
	}

	// clean up
	free(D);
	free(delta);
	free(gamma);
}


//--------------------------------------------------------------------------------------
// Fit a cubic spline to the control points
//--------------------------------------------------------------------------------------
void SplineInterpolate(int n, const ControlPoint *cpoints, XMVECTOR *coords)
{
	assert(cpoints[0].isoval == 0);

	SplineCoeffs *C = (SplineCoeffs *)malloc(n*sizeof(SplineCoeffs));
	CalculateSplineCoeffs(n, cpoints, C);

	for (int i = 0; i < n; i++)
	{
		int num_steps = cpoints[i+1].isoval - cpoints[i].isoval;

		for (int j = 0; j < num_steps; j++)
		{
			float s = j / (float)num_steps;
			coords[cpoints[i].isoval + j] = GetPointOnSpline(&C[i], s);
		}
	}

	free(C);
}



//void SplineTest()
//{
//	const int n = 8;
//	const int max_iso = 256;
//
//	ControlPoint *cpoints = (ControlPoint *)malloc((n+1) * sizeof(ControlPoint));
//	for (int i = 0; i <= n; i++)
//	{
//		cpoints[i].coord = XMVectorSet((float)(i*max_iso/n), 1, (float)(max_iso - i*max_iso/n), 0);
//		cpoints[i].isoval = i*max_iso/n;
//	}
//
//	XMVECTOR *coords = (XMVECTOR *)malloc(max_iso*sizeof(XMVECTOR));
//	SplineInterpolate(n, cpoints, coords);
//
//	for (int i = 0; i < max_iso; i++)
//	{
//		printf("(%g, %g, %g, %g)\n", XMVectorGetX(coords[i]), XMVectorGetY(coords[i]), XMVectorGetZ(coords[i]), XMVectorGetW(coords[i]));
//	}
//
//	free(coords);
//	free(cpoints);
//}
