//--------------------------------------------------------------------------------------
// DirectX/HLSL implementation by
// Christian B. Mendl, 2012
//
// Reference:
// http://graphicsrunner.blogspot.de/2009/01/volume-rendering-102-transfer-functions.html
//--------------------------------------------------------------------------------------

#include "gradient_filter.h"
#include <d3d11.h>
#include <dxerr.h>
#include <xnamath.h>
#include <io.h>
#include "util.h"


//--------------------------------------------------------------------------------------
// Constructor
//--------------------------------------------------------------------------------------
GradientFilter::GradientFilter()
{
	ZeroMemory(this, sizeof(*this));
}


HRESULT GradientFilter::CreateResources(ID3D11Device *pd3dDevice, UINT iWidth, UINT iHeight, UINT iDepth, wchar_t *fileName)
{
	HRESULT hr;

	// Store dimensions
	m_iWidth = iWidth;
	m_iHeight = iHeight;
	m_iDepth = iDepth;

	// Gradient compute shader
	{
		// Compile and create the shader
		ID3DBlob* pBlob = NULL;
		V_RETURN(CompileShaderFromFile(L"../shaders/gradient.hlsl", "ComputeGradient", "cs_5_0", &pBlob));
		V_RETURN(pd3dDevice->CreateComputeShader(pBlob->GetBufferPointer(), pBlob->GetBufferSize(), NULL, &m_pCSGradient));
	}

	// Filter compute shader
	{
		// Compile and create the shader
		ID3DBlob* pBlob = NULL;
		V_RETURN(CompileShaderFromFile(L"../shaders/filter.hlsl", "Filter", "cs_5_0", &pBlob));
		V_RETURN(pd3dDevice->CreateComputeShader(pBlob->GetBufferPointer(), pBlob->GetBufferSize(), NULL, &m_pCSFilter));
	}


	// Load the voxel volume data
	BYTE *buffer = NULL;
	{
		HANDLE hFile = CreateFileW(fileName,  GENERIC_READ, 0, NULL, OPEN_EXISTING, OPEN_EXISTING, NULL);
		if (hFile == INVALID_HANDLE_VALUE) {
			MessageBox(NULL, L"Opening volume data file failed.", L"Error", MB_ICONERROR | MB_OK);
			return E_FAIL;
		}

		buffer = (BYTE *)malloc(m_iWidth*m_iHeight*m_iDepth * sizeof(BYTE));

		DWORD numberOfBytesRead = 0;
		if (ReadFile(hFile, buffer, m_iWidth*m_iHeight*m_iDepth, &numberOfBytesRead, NULL) == 0)
		{
			MessageBox(NULL, L"Reading volume data failed.", L"Error", MB_ICONERROR | MB_OK);
			return E_FAIL;
		}
		if (numberOfBytesRead != m_iWidth*m_iHeight*m_iDepth)
		{
			MessageBox(NULL, L"Number of read bytes does not match.", L"Error", MB_ICONERROR | MB_OK);
			return E_FAIL;
		}

		CloseHandle(hFile);
	}

	// Create voxel volume texture
	{
		D3D11_TEXTURE3D_DESC descTex;
		ZeroMemory(&descTex, sizeof(descTex));
		descTex.Width  = m_iWidth;
		descTex.Height = m_iHeight;
		descTex.Depth  = m_iDepth;
		descTex.MipLevels = 1;
		descTex.Format = DXGI_FORMAT_R8_UNORM;
		descTex.Usage = D3D11_USAGE_DEFAULT;
		descTex.BindFlags = D3D11_BIND_SHADER_RESOURCE;
		descTex.CPUAccessFlags = 0;
		// Initial data
		D3D11_SUBRESOURCE_DATA initData;
		ZeroMemory(&initData, sizeof(initData));
		initData.pSysMem = buffer;
		initData.SysMemPitch = m_iWidth;
		initData.SysMemSlicePitch = m_iWidth*m_iHeight;
		// Create texture
		V_RETURN(pd3dDevice->CreateTexture3D(&descTex, &initData, &m_pTexVolume));
		free(buffer);

		// Create a resource view of the texture
		V_RETURN(pd3dDevice->CreateShaderResourceView(m_pTexVolume, NULL, &m_pTexVolumeRV));
	}


	// Create intermediate volume gradient and voxel texture
	{
		D3D11_TEXTURE3D_DESC descTex;
		ZeroMemory(&descTex, sizeof(descTex));
		descTex.Width  = m_iWidth;
		descTex.Height = m_iHeight;
		descTex.Depth  = m_iDepth;
		descTex.MipLevels = 1;
		descTex.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;	// store gradient in xyz and voxel instensity in w (alpha)
		descTex.Usage = D3D11_USAGE_DEFAULT;
		descTex.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS;
		descTex.CPUAccessFlags = 0;
		// Create texture
		V_RETURN(pd3dDevice->CreateTexture3D(&descTex, NULL, &m_pTexGradientVoxel));

		// Create a resource view of the texture
		V_RETURN(pd3dDevice->CreateShaderResourceView(m_pTexGradientVoxel, NULL, &m_pTexGradientVoxelRV));

		// Create an unorder access view of the texture
		//D3D11_UNORDERED_ACCESS_VIEW_DESC descView;
		//ZeroMemory(&descView, sizeof(descView));
		//descView.ViewDimension = D3D11_UAV_DIMENSION_TEXTURE3D;
		//descView.Format = DXGI_FORMAT_UNKNOWN;
		//descView.Texture3D.WSize = m_iDepth;
		V_RETURN(pd3dDevice->CreateUnorderedAccessView(m_pTexGradientVoxel, NULL, &m_pTexGradientVoxelUAV));
	}


	// Filtered volume gradient and voxel texture (final output)
	{
		// Create gradient texture
		D3D11_TEXTURE3D_DESC descTex;
		ZeroMemory(&descTex, sizeof(descTex));
		descTex.Width  = m_iWidth;
		descTex.Height = m_iHeight;
		descTex.Depth  = m_iDepth;
		descTex.MipLevels = 1;
		descTex.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;	// store gradient in xyz and voxel instensity in w (alpha)
		descTex.Usage = D3D11_USAGE_DEFAULT;
		descTex.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS;
		descTex.CPUAccessFlags = 0;
		// Create texture
		V_RETURN(pd3dDevice->CreateTexture3D(&descTex, NULL, &m_pTexFilteredGradientVoxel));

		// Create a resource view of the texture
		V_RETURN(pd3dDevice->CreateShaderResourceView(m_pTexFilteredGradientVoxel, NULL, &m_pTexFilteredGradientVoxelRV));

		// Create an unorder access view of the texture
		D3D11_UNORDERED_ACCESS_VIEW_DESC descView;
		ZeroMemory(&descView, sizeof(descView));
		descView.ViewDimension = D3D11_UAV_DIMENSION_TEXTURE3D;
		descView.Format = DXGI_FORMAT_UNKNOWN;
		descView.Texture3D.WSize = m_iDepth;
		V_RETURN(pd3dDevice->CreateUnorderedAccessView(m_pTexFilteredGradientVoxel, &descView, &m_pTexFilteredGradientVoxelUAV));
	}


	return S_OK;
}


//--------------------------------------------------------------------------------------
// Process the volume voxel data: compute and smooth gradients
//--------------------------------------------------------------------------------------
void GradientFilter::ProcessVolume(ID3D11DeviceContext *pImmediateContext)
{
	// Compute gradients
	{
		// Set the gradient compute shader
		pImmediateContext->CSSetShader(m_pCSGradient, NULL, 0);

		// Set input and output for the shader
		pImmediateContext->CSSetShaderResources(0, 1, &m_pTexVolumeRV);
		UINT initialCounts = 0;
		pImmediateContext->CSSetUnorderedAccessViews(0, 1, &m_pTexGradientVoxelUAV, &initialCounts);

		// Calculate gradients; shader uses [numthreads(8, 8, 8)]
		pImmediateContext->Dispatch(m_iWidth/8, m_iHeight/8, m_iDepth/8);

		// Un-bind shader resource view
		ID3D11ShaderResourceView *nullRV[] = { NULL };
		pImmediateContext->CSSetShaderResources(0, 1, nullRV);
		// Un-bind UAV
		ID3D11UnorderedAccessView *nullUAV[] = { NULL };
		pImmediateContext->CSSetUnorderedAccessViews(0, 1, nullUAV, &initialCounts);

		// Un-set the compute shader
		pImmediateContext->CSSetShader(NULL, NULL, 0);
	}

	/*// Dump calculated texture to disk for debugging
	{
		// Need copy for CPU access
		D3D11_TEXTURE3D_DESC descTex;
		ZeroMemory(&descTex, sizeof(descTex));
		descTex.Width  = m_iWidth;
		descTex.Height = m_iHeight;
		descTex.Depth  = m_iDepth;
		descTex.MipLevels = 1;
		descTex.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;	// store gradient in xyz and voxel instensity in w (alpha)
		descTex.Usage = D3D11_USAGE_STAGING;//D3D11_USAGE_DEFAULT;
		descTex.BindFlags = 0;//D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS;
		descTex.CPUAccessFlags = D3D11_CPU_ACCESS_READ;
		// Create texture
		ID3D11Texture3D *pTexGradientVoxelCopy = NULL;
		V_RETURN(pd3dDevice->CreateTexture3D(&descTex, NULL, &pTexGradientVoxelCopy));

		// Create a resource view of the texture
		//V_RETURN(pd3dDevice->CreateShaderResourceView(g_pTexFilteredGradientVoxel, NULL, &g_pTexFilteredGradientVoxelRV));

		// Copy texture for CPU access
		pImmediateContext->CopyResource(pTexGradientVoxelCopy, m_pTexGradientVoxel);

		// Open output (dump) file
		FILE *fd = NULL;
		if (fopen_s(&fd, "../media/tex_gradient_voxel_dump.m", "w") != 0)
		{
			wprintf(L"Could not open for writing gradient/isovalue texture.\n");
			return E_FAIL;
		}

		// Map texture
		D3D11_MAPPED_SUBRESOURCE mappedResource;
		V_RETURN(pImmediateContext->Map(pTexGradientVoxelCopy, 0, D3D11_MAP_READ, 0, &mappedResource));

		// corresponding to DXGI_FORMAT_R32G32B32A32_FLOAT
		XMFLOAT4 *data = (XMFLOAT4 *)mappedResource.pData;

		// output text format
		fprintf(fd, "{\n");
		for (int i = 150; i < 160; i++)//m_iDepth; i++)
		{
			fprintf(fd, "\t{\n");
			for (int j = 150; j < 160; j++)//m_iHeight; j++)
			{
				fprintf(fd, "\t\t{ ");
				for (int k = 150; k < 160; k++)//m_iWidth; k++)
				{
					XMFLOAT4 *curdata = &data[(i*m_iWidth + j)*m_iHeight + k];
					fprintf(fd, "{%g, %g, %g, %g}, ", curdata->x, curdata->y, curdata->z, curdata->w);
				}
				fseek(fd, -2, SEEK_CUR);
				fprintf(fd, " },\n");
			}
			fseek(fd, -3, SEEK_CUR);
			fprintf(fd, "\n\t},\n");
		}
		fseek(fd, -3, SEEK_CUR);
		fprintf(fd, "\n}\n");

		pImmediateContext->Unmap(pTexGradientVoxelCopy, 0);
		fclose(fd);

		pTexGradientVoxelCopy->Release();
	}*/


	// Filter gradients
	{
		// Set the gradient compute shader
		pImmediateContext->CSSetShader(m_pCSFilter, NULL, 0);

		// Set input and output for the shader
		pImmediateContext->CSSetShaderResources(0, 1, &m_pTexGradientVoxelRV);
		UINT initialCounts = 0;
		pImmediateContext->CSSetUnorderedAccessViews(0, 1, &m_pTexFilteredGradientVoxelUAV, &initialCounts);

		// Filter gradients; shader uses [numthreads(8, 8, 8)]
		pImmediateContext->Dispatch(m_iWidth/8, m_iHeight/8, m_iDepth/8);

		// Un-bind shader resource view
		ID3D11ShaderResourceView *nullRV[] = { NULL };
		pImmediateContext->CSSetShaderResources(0, 1, nullRV);
		// Un-bind UAV
		ID3D11UnorderedAccessView *nullUAV[] = { NULL };
		pImmediateContext->CSSetUnorderedAccessViews(0, 1, nullUAV, &initialCounts);

		// Un-set the compute shader
		pImmediateContext->CSSetShader(NULL, NULL, 0);
	}
}


//--------------------------------------------------------------------------------------
// Return resource view of the computed texture;
// caller must 'Release' resource after using it
//--------------------------------------------------------------------------------------
void GradientFilter::GetTextureRV(ID3D11ShaderResourceView **ppTexRV)
{
	(*ppTexRV) = m_pTexFilteredGradientVoxelRV;
	(*ppTexRV)->AddRef();
}


//--------------------------------------------------------------------------------------
// Release resources
//--------------------------------------------------------------------------------------
void GradientFilter::Cleanup()
{
	SAVE_RELEASE(m_pTexFilteredGradientVoxelUAV);
	SAVE_RELEASE(m_pTexFilteredGradientVoxelRV);
	SAVE_RELEASE(m_pTexFilteredGradientVoxel);

	SAVE_RELEASE(m_pTexGradientVoxelUAV);
	SAVE_RELEASE(m_pTexGradientVoxelRV);
	SAVE_RELEASE(m_pTexGradientVoxel);

	SAVE_RELEASE(m_pTexVolumeRV);
	SAVE_RELEASE(m_pTexVolume);

	SAVE_RELEASE(m_pCSFilter);
	SAVE_RELEASE(m_pCSGradient);
}
