//--------------------------------------------------------------------------------------
// DirectX/HLSL implementation by
// Christian B. Mendl, 2012
//
// Reference:
// http://graphicsrunner.blogspot.de/2009/01/volume-rendering-102-transfer-functions.html
//--------------------------------------------------------------------------------------

#include "model_position.h"
#include <d3d11.h>
#include <xnamath.h>
#include <dxerr.h>
#include "util.h"


//--------------------------------------------------------------------------------------
// Constant buffer structure
//--------------------------------------------------------------------------------------

struct CBEveryFrame
{
	XMMATRIX mWVP;
};


//--------------------------------------------------------------------------------------
// Constructor
//--------------------------------------------------------------------------------------
ModelPosition::ModelPosition()
{
	ZeroMemory(this, sizeof(*this));
}


//--------------------------------------------------------------------------------------
// Create resources
//--------------------------------------------------------------------------------------
HRESULT ModelPosition::CreateResources(ID3D11Device *pd3dDevice, UINT iWindowWidth, UINT iWindowHeight)
{
	HRESULT hr;


	// Model position shaders and input layout
	{
		// Compile and create the vertex shader
		ID3DBlob* pBlob = NULL;
		V_RETURN(CompileShaderFromFile(L"../shaders/model_position.hlsl", "ModelPositionVS", "vs_4_0", &pBlob));
		V_RETURN(pd3dDevice->CreateVertexShader(pBlob->GetBufferPointer(), pBlob->GetBufferSize(), NULL, &m_pVSModelPosition));

		// Define the input layout
		D3D11_INPUT_ELEMENT_DESC layout[] =
		{
			{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0,  0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
		};
		UINT numElements = ARRAYSIZE(layout);

		// Create the input layout
		// Note that 'pBlob' is inherited from the above vertex shader
		V_RETURN(pd3dDevice->CreateInputLayout(layout, numElements, pBlob->GetBufferPointer(), pBlob->GetBufferSize(), &m_pVLModelPosition));
		pBlob->Release();
	
		// Compile and create the pixel shader
		pBlob = NULL;
		V_RETURN(CompileShaderFromFile(L"../shaders/model_position.hlsl", "ModelPositionPS", "ps_4_0", &pBlob));
		V_RETURN(pd3dDevice->CreatePixelShader(pBlob->GetBufferPointer(), pBlob->GetBufferSize(), NULL, &m_pPSModelPosition));
		pBlob->Release();
	}


	// Create the 2D position render target textures
	{
		D3D11_TEXTURE2D_DESC descTex;
		ZeroMemory(&descTex, sizeof(descTex));
		descTex.ArraySize = 1;
		descTex.BindFlags = D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE;
		descTex.Usage = D3D11_USAGE_DEFAULT;
		descTex.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
		descTex.Width = iWindowWidth;
		descTex.Height = iWindowHeight;
		descTex.MipLevels = 1;
		descTex.SampleDesc.Count = 1;
		descTex.CPUAccessFlags = 0;

		// Front and back side
		for (int i = 0; i < 2; i++)
		{
			V_RETURN(pd3dDevice->CreateTexture2D(&descTex, NULL, &m_pTexPosition[i]));

			// Create resource view
			V_RETURN(pd3dDevice->CreateShaderResourceView(m_pTexPosition[i], NULL, &m_pTexPositionRV[i]));

			// Create render target view
			V_RETURN(pd3dDevice->CreateRenderTargetView(m_pTexPosition[i], NULL, &m_pTexPositionRTV[i]));
		}
	}

	// Rasterizer states: front and back face culling
	{
		// Back face culling
		D3D11_RASTERIZER_DESC rasterizerDesc;
		ZeroMemory(&rasterizerDesc, sizeof(rasterizerDesc));
		rasterizerDesc.FillMode = D3D11_FILL_SOLID;
		rasterizerDesc.CullMode = D3D11_CULL_BACK;
		rasterizerDesc.DepthClipEnable = true;
		V_RETURN(pd3dDevice->CreateRasterizerState(&rasterizerDesc, &m_pRasterizerStateCullBack));

		// Front face culling
		ZeroMemory(&rasterizerDesc, sizeof(rasterizerDesc));
		rasterizerDesc.FillMode = D3D11_FILL_SOLID;
		rasterizerDesc.CullMode = D3D11_CULL_FRONT;
		rasterizerDesc.DepthClipEnable = true;
		V_RETURN(pd3dDevice->CreateRasterizerState(&rasterizerDesc, &m_pRasterizerStateCullFront));
	}

	// Constant
	{
		D3D11_BUFFER_DESC bd;
		ZeroMemory(&bd, sizeof(bd));
		bd.Usage = D3D11_USAGE_DEFAULT;
		bd.ByteWidth = sizeof(CBEveryFrame);
		bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
		bd.CPUAccessFlags = 0;
		V_RETURN(pd3dDevice->CreateBuffer(&bd, NULL, &m_pCBEveryFrame));
	}


	return S_OK;
}


//--------------------------------------------------------------------------------------
// Render cube positions to textures
//--------------------------------------------------------------------------------------
void ModelPosition::Render(ID3D11DeviceContext *pImmediateContext, const XMMATRIX *pWVP)
{
	// Set the input layout
	pImmediateContext->IASetInputLayout(m_pVLModelPosition);

	// Update constant buffer
	CBEveryFrame cb;
	cb.mWVP = *pWVP;
	pImmediateContext->UpdateSubresource(m_pCBEveryFrame, 0, NULL, &cb, 0, 0);

	// Set the vertex shader
	pImmediateContext->VSSetShader(m_pVSModelPosition, NULL, 0);
	pImmediateContext->VSSetConstantBuffers(0, 1, &m_pCBEveryFrame);

	// Set the pixel shader
	pImmediateContext->PSSetShader(m_pPSModelPosition, NULL, 0);

	// Clear render targets
	float clearColor[4] = { 0.f, 0.f, 0.f, 1.f };	// red, green, blue, alpha
	pImmediateContext->ClearRenderTargetView(m_pTexPositionRTV[0], clearColor);
	pImmediateContext->ClearRenderTargetView(m_pTexPositionRTV[1], clearColor);

	// Front-face culling
	pImmediateContext->RSSetState(m_pRasterizerStateCullFront);
	pImmediateContext->OMSetRenderTargets(1, &m_pTexPositionRTV[1], NULL);
	// Draw cube back faces
	pImmediateContext->DrawIndexed(36, 0, 0);

	// Back-face culling
	pImmediateContext->RSSetState(m_pRasterizerStateCullBack);
	pImmediateContext->OMSetRenderTargets(1, &m_pTexPositionRTV[0], NULL);
	// Draw cube front faces
	pImmediateContext->DrawIndexed(36, 0, 0);
}


//--------------------------------------------------------------------------------------
// Return resource views of the computed position textures;
// caller must 'Release' resources after using them
//--------------------------------------------------------------------------------------
void ModelPosition::GetTextureRV(ID3D11ShaderResourceView *ppTexRV[2])
{
	ppTexRV[0] = m_pTexPositionRV[0];
	ppTexRV[0]->AddRef();

	ppTexRV[1] = m_pTexPositionRV[1];
	ppTexRV[1]->AddRef();
}



//--------------------------------------------------------------------------------------
// Release resources
//--------------------------------------------------------------------------------------
void ModelPosition::Cleanup()
{
	SAVE_RELEASE(m_pCBEveryFrame);

	SAVE_RELEASE(m_pRasterizerStateCullFront);
	SAVE_RELEASE(m_pRasterizerStateCullBack);

	for (int i = 0; i < 2; i++)
	{
		SAVE_RELEASE(m_pTexPositionRTV[i]);
		SAVE_RELEASE(m_pTexPositionRV[i]);
		SAVE_RELEASE(m_pTexPosition[i]);
	}

	SAVE_RELEASE(m_pPSModelPosition);
	SAVE_RELEASE(m_pVLModelPosition);
	SAVE_RELEASE(m_pVSModelPosition);
}
