//--------------------------------------------------------------------------------------
// DirectX/HLSL implementation by
// Christian B. Mendl, 2012
//
// Reference:
// http://graphicsrunner.blogspot.de/2009/01/volume-rendering-102-transfer-functions.html
//--------------------------------------------------------------------------------------

#include "raycast.h"
#include <d3d11.h>
#include <xnamath.h>
#include <dxerr.h>
#include <stdio.h>
#include <io.h>
#include "cubic_spline.h"
#include "util.h"


//--------------------------------------------------------------------------------------
// Constant buffer structures
//--------------------------------------------------------------------------------------

struct CBEveryFrameVS
{
	XMMATRIX mWVP;
};


struct CBImmutablePS
{
	float fInvWindowSize[2];
	float dummy[2];		// byte width must be a multiple of 16
};

struct CBEveryFramePS
{
	XMVECTOR vLightPos;		// in model coordinates
};



//--------------------------------------------------------------------------------------
// Constructor
//--------------------------------------------------------------------------------------
RayCast::RayCast()
{
	ZeroMemory(this, sizeof(*this));
}


//--------------------------------------------------------------------------------------
// Create resources
//--------------------------------------------------------------------------------------
HRESULT RayCast::CreateResources(ID3D11Device *pd3dDevice, UINT iWindowWidth, UINT iWindowHeight)
{
	HRESULT hr;

	// Ray-casting shaders and input layout
	{
		// Compile and create the vertex shader
		ID3DBlob* pBlob = NULL;
		V_RETURN(CompileShaderFromFile(L"../shaders/raycast.hlsl", "RayCastVS", "vs_4_0", &pBlob));
		V_RETURN(pd3dDevice->CreateVertexShader(pBlob->GetBufferPointer(), pBlob->GetBufferSize(), NULL, &m_pVSRayCast));

		// Define the input layout
		D3D11_INPUT_ELEMENT_DESC layout[] =
		{
			{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0,  0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
		};
		UINT numElements = ARRAYSIZE(layout);

		// Create the input layout
		// Note that 'pBlob' is inherited from the above vertex shader
		V_RETURN(pd3dDevice->CreateInputLayout(layout, numElements, pBlob->GetBufferPointer(), pBlob->GetBufferSize(), &m_pVLRayCast));
		pBlob->Release();
	
		// Compile and create the pixel shaders
		pBlob = NULL;
		V_RETURN(CompileShaderFromFile(L"../shaders/raycast.hlsl", "RayCastPS", "ps_4_0", &pBlob));
		V_RETURN(pd3dDevice->CreatePixelShader(pBlob->GetBufferPointer(), pBlob->GetBufferSize(), NULL, &m_pPSRayCast));
		pBlob->Release();
	}

	// Create the transfer texture (color and transparency)
	V_RETURN(CreateTransferTexture(pd3dDevice));


	// Vertex shader constant buffer
	{
		D3D11_BUFFER_DESC bd;
		ZeroMemory(&bd, sizeof(bd));
		bd.Usage = D3D11_USAGE_DEFAULT;
		bd.ByteWidth = sizeof(CBEveryFrameVS);
		bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
		bd.CPUAccessFlags = 0;
		V_RETURN(pd3dDevice->CreateBuffer(&bd, NULL, &m_pCBEveryFrameVS));
	}

	// Pixel shader 'immutable' constant buffer
	{
		CBImmutablePS cb;
		cb.fInvWindowSize[0] = 1.f/iWindowWidth;
		cb.fInvWindowSize[1] = 1.f/iWindowHeight;

		D3D11_BUFFER_DESC bd;
		ZeroMemory(&bd, sizeof(bd));
		bd.Usage = D3D11_USAGE_IMMUTABLE;
		bd.ByteWidth = sizeof(CBImmutablePS);
		bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
		bd.CPUAccessFlags = 0;
		D3D11_SUBRESOURCE_DATA initData;
		ZeroMemory(&initData, sizeof(initData));
		initData.pSysMem = &cb;
		V_RETURN(pd3dDevice->CreateBuffer(&bd, &initData, &m_pCBImmutablePS));
	}

	// Pixel shader 'every frame' constant buffer
	{
		D3D11_BUFFER_DESC bd;
		ZeroMemory(&bd, sizeof(bd));
		bd.Usage = D3D11_USAGE_DEFAULT;
		bd.ByteWidth = sizeof(CBEveryFramePS);
		bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
		bd.CPUAccessFlags = 0;
		V_RETURN(pd3dDevice->CreateBuffer(&bd, NULL, &m_pCBEveryFramePS));
	}


	return S_OK;
}


//--------------------------------------------------------------------------------------
// Create the transfer function texture
// Reference: http://graphicsrunner.blogspot.de/2009/01/volume-rendering-102-transfer-functions.html
//--------------------------------------------------------------------------------------
HRESULT RayCast::CreateTransferTexture(ID3D11Device *pd3dDevice)
{
	HRESULT hr;


	// Setup the control points for the color and alpha knots
	// note: you must make sure that you have at least color/alpha
	// points for 0 and 256, else the interpolation will fail

	ControlPoint colorKnots[4] = {
		{ XMVectorSet(0.91f, 0.7f, 0.61f, 1.f),   0 },	// Skin
		{ XMVectorSet(0.91f, 0.7f, 0.61f, 1.f),  80 },	// Skin
		{ XMVectorSet(1.f,   1.f,  0.85f, 1.f),  82 },	// Bone
		{ XMVectorSet(1.f,   1.f,  0.85f, 1.f), 256 }	// Bone
	};

	// Skin ranges from about 40 - 70
	// Bone ranges from about 70 - 255
	ControlPoint alphaKnots[7] = {
		{ XMVectorSet(0.f, 0.f, 0.f, 0.f),     0 },
		{ XMVectorSet(0.f, 0.f, 0.f, 0.f),    40 },
		{ XMVectorSet(0.f, 0.f, 0.f, 0.2f),   60 },
		{ XMVectorSet(0.f, 0.f, 0.f, 0.05f),  63 },
		{ XMVectorSet(0.f, 0.f, 0.f, 0.f),    80 },
		{ XMVectorSet(0.f, 0.f, 0.f, 0.9f),   82 },
		{ XMVectorSet(0.f, 0.f, 0.f, 1.f),   256 },
	};

	const int max_iso = 256;

	XMVECTOR *colorCoords = (XMVECTOR *)malloc(max_iso*sizeof(XMVECTOR));
	SplineInterpolate(ARRAYSIZE(colorKnots)-1, colorKnots, colorCoords);

	XMVECTOR *alphaCoords = (XMVECTOR *)malloc(max_iso*sizeof(XMVECTOR));
	SplineInterpolate(ARRAYSIZE(alphaKnots)-1, alphaKnots, alphaCoords);

	// Interleave color and alpha
	XMVECTOR *coords = (XMVECTOR *)malloc(max_iso*sizeof(XMVECTOR));
	for (int i = 0; i < max_iso; i++) {
		coords[i] = XMVectorClamp(XMVectorSet(XMVectorGetX(colorCoords[i]), XMVectorGetY(colorCoords[i]), XMVectorGetZ(colorCoords[i]), XMVectorGetW(alphaCoords[i])),
			XMVectorSet(0, 0, 0, 0), XMVectorSet(1, 1, 1, 1));
	}

	// Create and fill transfer texture
	D3D11_TEXTURE1D_DESC descTex;
	ZeroMemory(&descTex, sizeof(descTex));
	descTex.ArraySize = 1;
	descTex.BindFlags = D3D11_BIND_SHADER_RESOURCE;
	descTex.Usage = D3D11_USAGE_DEFAULT;
	descTex.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
	descTex.Width = max_iso;
	descTex.MipLevels = 1;
	descTex.CPUAccessFlags = 0;
	// Initial data
	D3D11_SUBRESOURCE_DATA initData;
	ZeroMemory(&initData, sizeof(initData));
	initData.pSysMem = coords;
	V_RETURN(pd3dDevice->CreateTexture1D(&descTex, &initData, &m_pTexTransfer));

	// Create resource view
	pd3dDevice->CreateShaderResourceView(m_pTexTransfer, NULL, &m_pTexTransferRV);


	// Save data to file for debugging

	FILE *fd = NULL;
	if (fopen_s(&fd, "../media/tex_transfer_dump.m", "w") != 0)
	{
		wprintf(L"Could not open file for writing transfer texture data.\n");
		return E_FAIL;
	}
	fprintf(fd, "{ ");
	for (UINT i = 0; i < descTex.Width; i++)
	{
		fprintf(fd, "{%g, %g, %g, %g}, ", XMVectorGetX(coords[i]), XMVectorGetY(coords[i]), XMVectorGetZ(coords[i]), XMVectorGetW(coords[i]));
	}
	fseek(fd, -2, SEEK_CUR);
	fprintf(fd, " }");
	fclose(fd);

	free(coords);
	free(alphaCoords);
	free(colorCoords);

	return S_OK;
}


//--------------------------------------------------------------------------------------
// Ray-casting
//--------------------------------------------------------------------------------------
void RayCast::Render(ID3D11DeviceContext *pImmediateContext, ID3D11ShaderResourceView *pTexGradientVoxelRV, ID3D11ShaderResourceView *pTexPositionRV[2], const XMMATRIX *pWVP, const XMVECTOR *vLightPos)
{
	// Set the input layout
	pImmediateContext->IASetInputLayout(m_pVLRayCast);

	// Update constant buffer
	CBEveryFrameVS cbVS;
	cbVS.mWVP = *pWVP;
	pImmediateContext->UpdateSubresource(m_pCBEveryFrameVS, 0, NULL, &cbVS, 0, 0);

	// Set the vertex shader
	pImmediateContext->VSSetShader(m_pVSRayCast, NULL, 0);
	pImmediateContext->VSSetConstantBuffers(0, 1, &m_pCBEveryFrameVS);

	// Update constant buffer
	CBEveryFramePS cbPS;
	cbPS.vLightPos = *vLightPos;
	pImmediateContext->UpdateSubresource(m_pCBEveryFramePS, 0, NULL, &cbPS, 0, 0);

	// Set the pixel shader
	pImmediateContext->PSSetShader(m_pPSRayCast, NULL, 0);
	pImmediateContext->PSSetConstantBuffers(0, 1, &m_pCBImmutablePS);
	pImmediateContext->PSSetConstantBuffers(1, 1, &m_pCBEveryFramePS);

	// Set textures
	pImmediateContext->PSSetShaderResources(0, 1, &pTexGradientVoxelRV);
	pImmediateContext->PSSetShaderResources(1, 2, pTexPositionRV);
	pImmediateContext->PSSetShaderResources(3, 1, &m_pTexTransferRV);

	// Draw the cube
	pImmediateContext->DrawIndexed(36, 0, 0);

	// Un-bind textures
	ID3D11ShaderResourceView *nullRV[4] = { NULL, NULL, NULL, NULL };
	pImmediateContext->PSSetShaderResources(0, 4, nullRV);
}


//--------------------------------------------------------------------------------------
// Release resources
//--------------------------------------------------------------------------------------
void RayCast::Cleanup()
{
	SAVE_RELEASE(m_pCBEveryFramePS);
	SAVE_RELEASE(m_pCBImmutablePS);
	SAVE_RELEASE(m_pCBEveryFrameVS);

	SAVE_RELEASE(m_pTexTransferRV);
	SAVE_RELEASE(m_pTexTransfer);

	SAVE_RELEASE(m_pPSRayCast);
	SAVE_RELEASE(m_pVLRayCast);
	SAVE_RELEASE(m_pVSRayCast);
}
