//--------------------------------------------------------------------------------------
// Volume rendering by ray-casting
//
// DirectX/HLSL implementation by
// Christian B. Mendl, 2012
//
// Reference:
// http://graphicsrunner.blogspot.de/2009/01/volume-rendering-102-transfer-functions.html
//--------------------------------------------------------------------------------------

#include <d3d11.h>
#include <xnamath.h>
#include <dxerr.h>
#include <crtdbg.h>
#include <stdio.h>
#include "model_position.h"
#include "gradient_filter.h"
#include "raycast.h"
#include "util.h"


//--------------------------------------------------------------------------------------
// Global Variables
//--------------------------------------------------------------------------------------

D3D_DRIVER_TYPE						g_driverType = D3D_DRIVER_TYPE_NULL;
D3D_FEATURE_LEVEL					g_featureLevel;
ID3D11Device*						g_pd3dDevice = NULL;
ID3D11DeviceContext*				g_pImmediateContext = NULL;
DXGI_ADAPTER_DESC1					g_sAdapterDesc;

HINSTANCE							g_hInst = NULL;
HWND								g_hWnd = NULL;
IDXGISwapChain*						g_pSwapChain = NULL;
ID3D11RenderTargetView*				g_pRenderTargetView = NULL;
UINT								g_iWindowWidth = 0;
UINT								g_iWindowHeight = 0;

// Rendering modules
ModelPosition						g_cModelPosition;
GradientFilter						g_cGradientFilter;
RayCast								g_cRayCast;

// Default linear sampler
ID3D11SamplerState*					g_pSamplerLinear = NULL;

// Cube vertex and index buffers
ID3D11Buffer*						g_pVertexBufferCube = NULL;
ID3D11Buffer*						g_pIndexBufferCube = NULL;

// Matrices
XMMATRIX							g_mViewProjection;
const XMVECTOR						g_vLightPos = XMVectorSet(0, 4, -1, 0);		// in world coordinates

// Timing variables
__int64								g_iFreq = 0;
__int64								g_iStartTime = 0;
DWORD								g_dwFrameCount = 0;
float								g_fTime = 0;
float								g_fLastTime = 0;



//--------------------------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------------------------
HRESULT InitWindow(HINSTANCE hInstance, int nCmdShow);
HRESULT InitDevice();
void CleanupDevice();
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
void Render();


//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow)
{
	HRESULT hr;

	// Enable run-time memory check for debug builds.
	#if defined(DEBUG) | defined(_DEBUG)
	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
	#endif

	//UNREFERENCED_PARAMETER(hPrevInstance);
	//UNREFERENCED_PARAMETER(lpCmdLine);

	V_RETURN(InitWindow(hInstance, nCmdShow));

	SetupConsole();

	// Timing
	// Read the frequency of the system clock
	QueryPerformanceFrequency((LARGE_INTEGER *)&g_iFreq);
	// Obtain the start time
	QueryPerformanceCounter((LARGE_INTEGER *)&g_iStartTime);

	V_RETURN(InitDevice());

	// Main message loop
	MSG msg = { 0 };
	while(WM_QUIT != msg.message)
	{
		if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
		{
			// Current time
			__int64 iTime;
			QueryPerformanceCounter((LARGE_INTEGER *)&iTime);
			iTime -= g_iStartTime;
			g_fTime = (float)iTime / g_iFreq;

			// Render the scene
			Render();

			// Increment frame counter
			g_dwFrameCount++;
			// FPS display
			if (g_fTime - g_fLastTime > 1)
			{
				// Calculate updated frames per second
				wprintf(L"FPS: %.4g\n", g_dwFrameCount / (g_fTime - g_fLastTime));

				g_fLastTime = g_fTime;
				g_dwFrameCount = 0;
			}
		}
	}

	CleanupDevice();

	return (int)msg.wParam;
}


//--------------------------------------------------------------------------------------
// Register class and create window
//--------------------------------------------------------------------------------------
HRESULT InitWindow(HINSTANCE hInstance, int nCmdShow)
{
	// Register class
	WNDCLASSEX wcex;
	ZeroMemory(&wcex, sizeof(wcex));
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.style = CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc = WndProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hInstance;
	//wcex.hIcon = LoadIcon(hInstance, (LPCTSTR)IDI_IRADON);
	//wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	//wcex.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = L"VolumeRendererClass";
	//wcex.hIconSm = LoadIcon(wcex.hInstance, (LPCTSTR)IDI_IRADON);
	if(!RegisterClassEx(&wcex))
		return E_FAIL;

	// Create window
	g_hInst = hInstance;
	RECT rc = { 0, 0, 512, 512 };
	AdjustWindowRect(&rc, WS_OVERLAPPEDWINDOW, FALSE);
	g_hWnd = CreateWindow(L"VolumeRendererClass", L"Volume Ray-Casting Renderer, (c) Christian B. Mendl, 2012", WS_OVERLAPPEDWINDOW,
						  CW_USEDEFAULT, CW_USEDEFAULT, rc.right - rc.left, rc.bottom - rc.top, NULL, NULL, hInstance, NULL);
	if(!g_hWnd)
		return E_FAIL;

	ShowWindow(g_hWnd, nCmdShow);

	return S_OK;
}



//--------------------------------------------------------------------------------------
// Create Direct3D device and swap chain
//--------------------------------------------------------------------------------------
HRESULT InitDevice()
{
	HRESULT hr = S_OK;


	RECT rc;
	GetClientRect(g_hWnd, &rc);
	g_iWindowWidth = rc.right - rc.left;
	g_iWindowHeight = rc.bottom - rc.top;

	UINT createDeviceFlags = 0;
	#ifdef _DEBUG
	createDeviceFlags |= D3D11_CREATE_DEVICE_DEBUG;
	#endif

	// Enumerate driver types
	D3D_DRIVER_TYPE driverTypes[] =
	{
		D3D_DRIVER_TYPE_HARDWARE,
		D3D_DRIVER_TYPE_WARP,
		D3D_DRIVER_TYPE_REFERENCE,
	};
	UINT numDriverTypes = ARRAYSIZE(driverTypes);

	D3D_FEATURE_LEVEL featureLevels[] =
	{
		D3D_FEATURE_LEVEL_11_0,
		D3D_FEATURE_LEVEL_10_1,
		D3D_FEATURE_LEVEL_10_0,
	};
	UINT numFeatureLevels = ARRAYSIZE(featureLevels);

	// Fill the swap chain description
	DXGI_SWAP_CHAIN_DESC sd;
	ZeroMemory(&sd, sizeof(sd));
	sd.BufferCount = 1;
	sd.BufferDesc.Width = g_iWindowWidth;
	sd.BufferDesc.Height = g_iWindowHeight;
	sd.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
	sd.BufferDesc.RefreshRate.Numerator = 60;
	sd.BufferDesc.RefreshRate.Denominator = 1;
	sd.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
	sd.OutputWindow = g_hWnd;
	sd.SampleDesc.Count = 1;
	sd.SampleDesc.Quality = 0;
	sd.Windowed = TRUE;

	// Select an appropriate driver
	for (UINT driverTypeIndex = 0; driverTypeIndex < numDriverTypes; driverTypeIndex++)
	{
		g_driverType = driverTypes[driverTypeIndex];
		hr = D3D11CreateDeviceAndSwapChain(NULL, g_driverType, NULL, createDeviceFlags, featureLevels, numFeatureLevels,
											D3D11_SDK_VERSION, &sd, &g_pSwapChain, &g_pd3dDevice, &g_featureLevel, &g_pImmediateContext);
		if (SUCCEEDED(hr))
			break;
	}
	V_RETURN(hr);

	// Obtain adapter description
	if(g_driverType == D3D_DRIVER_TYPE_HARDWARE)
	{
		// Create a "DXGI factory" to enumerate adapters
		IDXGIFactory1 *pFactory;
		V_RETURN(CreateDXGIFactory1(__uuidof(IDXGIFactory1), (void **)&pFactory));

		// Obtain a description of the first adapter, which is (by default) used as rendering device
		IDXGIAdapter1 *pAdapter;
		V_RETURN(pFactory->EnumAdapters1(0, &pAdapter));
		pAdapter->GetDesc1(&g_sAdapterDesc);
		pAdapter->Release();

		pFactory->Release();
	}

	// Display device information
	const LPCWSTR g_strDriverType[] =		// Strings matching 'D3D_DRIVER_TYPE'
	{
		L"UNKNOWN",
		L"HARDWARE",
		L"REFERENCE",
		L"NULL",
		L"SOFTWARE",
		L"WARP",
	};
	wprintf (L"Driver type: %s\n", g_strDriverType[g_driverType]);
	if (g_driverType == D3D_DRIVER_TYPE_HARDWARE)
	{
		wprintf(L"%s\n", g_sAdapterDesc.Description);
	}

	// Create a render target view for the swap chain
	ID3D11Texture2D *pBackBuffer = NULL;
	V_RETURN(g_pSwapChain->GetBuffer(0, __uuidof(ID3D11Texture2D), (LPVOID *)&pBackBuffer));

	V_RETURN(g_pd3dDevice->CreateRenderTargetView(pBackBuffer, NULL, &g_pRenderTargetView));
	pBackBuffer->Release();

	g_pImmediateContext->OMSetRenderTargets(1, &g_pRenderTargetView, NULL);

	// Setup the viewport
	D3D11_VIEWPORT vp;
	vp.Width = (FLOAT)g_iWindowWidth;
	vp.Height = (FLOAT)g_iWindowHeight;
	vp.MinDepth = 0.f;
	vp.MaxDepth = 1.f;
	vp.TopLeftX = 0;
	vp.TopLeftY = 0;
	g_pImmediateContext->RSSetViewports(1, &vp);


	// Create module resources

	V_RETURN(g_cGradientFilter.CreateResources(g_pd3dDevice, 128, 256, 256, L"../media/VisMale.raw"/*L"../media/foot.raw"*/));

	V_RETURN(g_cModelPosition.CreateResources(g_pd3dDevice, g_iWindowWidth, g_iWindowHeight));

	V_RETURN(g_cRayCast.CreateResources(g_pd3dDevice, g_iWindowWidth, g_iWindowHeight));

	
	// Create the sampler state
	{
		D3D11_SAMPLER_DESC sampDesc;
		ZeroMemory(&sampDesc, sizeof(sampDesc));
		sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_BORDER;
		sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_BORDER;
		sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_BORDER;
		sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
		sampDesc.MinLOD = 0;
		sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
		V_RETURN(g_pd3dDevice->CreateSamplerState(&sampDesc, &g_pSamplerLinear));
	}

	// Set texture sampler
	g_pImmediateContext->PSSetSamplers(0, 1, &g_pSamplerLinear);


	// Create cube vertex and index buffers
	{
		XMFLOAT3 vertices[] =
		{
			XMFLOAT3(-1.f, -1.f, -1.f),
			XMFLOAT3(-1.f, -1.f,  1.f),
			XMFLOAT3(-1.f,  1.f, -1.f),
			XMFLOAT3(-1.f,  1.f,  1.f),
			XMFLOAT3( 1.f, -1.f, -1.f),
			XMFLOAT3( 1.f, -1.f,  1.f),
			XMFLOAT3( 1.f,  1.f, -1.f),
			XMFLOAT3( 1.f,  1.f,  1.f),
		};
		D3D11_BUFFER_DESC bd;
		ZeroMemory(&bd, sizeof(bd));
		bd.Usage = D3D11_USAGE_IMMUTABLE;
		bd.ByteWidth = sizeof(XMFLOAT3) * ARRAYSIZE(vertices);
		bd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
		bd.CPUAccessFlags = 0;
		D3D11_SUBRESOURCE_DATA initData;
		ZeroMemory(&initData, sizeof(initData));
		initData.pSysMem = vertices;
		V_RETURN(g_pd3dDevice->CreateBuffer(&bd, &initData, &g_pVertexBufferCube));

		// Create index buffer
		WORD indices[] =
		{
			0, 1, 2,
			2, 1, 3,

			0, 4, 1,
			1, 4, 5,

			0, 2, 4,
			4, 2, 6,

			1, 5, 3,
			3, 5, 7,

			2, 3, 6,
			6, 3, 7,

			5, 4, 7,
			7, 4, 6,
		};
		ZeroMemory(&bd, sizeof(bd));
		bd.Usage = D3D11_USAGE_IMMUTABLE;
		bd.ByteWidth = sizeof(WORD) * ARRAYSIZE(indices);
		bd.BindFlags = D3D11_BIND_INDEX_BUFFER;
		bd.CPUAccessFlags = 0;
		ZeroMemory(&initData, sizeof(initData));
		initData.pSysMem = indices;
		V_RETURN(g_pd3dDevice->CreateBuffer(&bd, &initData, &g_pIndexBufferCube));
	}


	// Matrices
	{
		// Initialize the view matrix
		XMVECTOR eye = XMVectorSet(0.f,  0.5f, -4.5f, 0.f);
		XMVECTOR at  = XMVectorSet(0.f,  0.f,   0.f,  0.f);
		XMVECTOR up  = XMVectorSet(0.f,  1.f,   0.f,  0.f);
		XMMATRIX mView = XMMatrixTranspose(XMMatrixLookAtLH(eye, at, up));

		// Initialize the projection matrix
		XMMATRIX mProjection = XMMatrixTranspose(XMMatrixPerspectiveFovLH(XM_PIDIV4, 1.f, 0.1f, 10.f));

		// View-projection matrix
		g_mViewProjection = XMMatrixMultiply(mProjection, mView);
	}


	// Pre-process the volume data

	printf("Pre-processing volume...\n");
	g_cGradientFilter.ProcessVolume(g_pImmediateContext);


	return S_OK;
}



//--------------------------------------------------------------------------------------
// Clean up
//--------------------------------------------------------------------------------------
void CleanupDevice()
{
	if(g_pImmediateContext) g_pImmediateContext->ClearState();

	SAVE_RELEASE(g_pSamplerLinear);

	SAVE_RELEASE(g_pIndexBufferCube);
	SAVE_RELEASE(g_pVertexBufferCube);

	g_cRayCast.Cleanup();
	g_cModelPosition.Cleanup();
	g_cGradientFilter.Cleanup();

	SAVE_RELEASE(g_pRenderTargetView);
	SAVE_RELEASE(g_pSwapChain);
	SAVE_RELEASE(g_pImmediateContext);
	SAVE_RELEASE(g_pd3dDevice);
}


//--------------------------------------------------------------------------------------
// Called every time the application receives a message
//--------------------------------------------------------------------------------------
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC hdc;

	switch(message)
	{
		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			EndPaint(hWnd, &ps);
			break;

		case WM_DESTROY:
			PostQuitMessage(0);
			break;

		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
	}

	return 0;
}


//--------------------------------------------------------------------------------------
// Render a frame
//--------------------------------------------------------------------------------------
void Render()
{
	// Clear the back buffer
	float clearColor[4] = { 0.f, 0.f, 0.f, 1.f };	// red, green, blue, alpha
	g_pImmediateContext->ClearRenderTargetView(g_pRenderTargetView, clearColor);


	// Set vertex buffer
	UINT stride = sizeof(XMFLOAT3);
	UINT offset = 0;
	g_pImmediateContext->IASetVertexBuffers(0, 1, &g_pVertexBufferCube, &stride, &offset);
	// Set index buffer
	g_pImmediateContext->IASetIndexBuffer(g_pIndexBufferCube, DXGI_FORMAT_R16_UINT, 0);

	// Set primitive topology
	g_pImmediateContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);


	// Rotate the cube around the origin
	XMMATRIX mWorld = XMMatrixRotationY(0.5f*XM_PIDIV4*g_fTime);
	// Light vector in model coordinates
	XMVECTOR vLightPosModel = XMVector4Transform(g_vLightPos, mWorld);	// for a rotation matrix, transposition is inverse
	// World-view-projection matrix
	XMMATRIX mWVP = XMMatrixMultiply(g_mViewProjection, mWorld);


	// Render model positions to texture
	g_cModelPosition.Render(g_pImmediateContext, &mWVP);


	// Ray-casting

	// Obtain texture resource view references
	ID3D11ShaderResourceView *pTexFilteredGradientVoxelRV = NULL;
	g_cGradientFilter.GetTextureRV(&pTexFilteredGradientVoxelRV);

	ID3D11ShaderResourceView *pTexPositionRV[2] = { NULL, NULL };
	g_cModelPosition.GetTextureRV(pTexPositionRV);

	// Render to standard render target
	g_pImmediateContext->OMSetRenderTargets(1, &g_pRenderTargetView, NULL);

	// Ray-casting rendering pass
	g_cRayCast.Render(g_pImmediateContext, pTexFilteredGradientVoxelRV, pTexPositionRV, &mWVP, &vLightPosModel);

	// Release texture references
	pTexFilteredGradientVoxelRV->Release();
	pTexPositionRV[1]->Release();
	pTexPositionRV[0]->Release();


	// Present the back buffer to the front buffer
	g_pSwapChain->Present(0, 0);
}
